
/*
   http://www.cb.uu.se/~cris/blog/index.php/archives/22

   The problem with these filters is that they respond very strongly to
   noise. Hence the introduction of filters like Prewitt and Sobel, that add
   some smoothing in the direction orthogonal to the derivative direction. Of
   course, these filters had to use a uniform or a triangular filter, instead
   of a good smoothing filter. (The triangular filter has better properties
   than the uniform filter, but does not suppress high frequencies as well as
   the Gaussian.) But even if we ignore the poor choice of smoothing filter,
   this type of derivative operator is in general a bad idea. Why? Because
   each of the components of a gradient vector computed with these filters
   has a different smoothing. Let me explain this. A filter like Prewitt is
   separable, equivalent of applying the [1,0,-1]/2 along one axis, and the
   [1,1,1]/3 filter along the other axis. So basically we are computing the
   partial derivative along the x-axis of an image that was smoothed with
   a 13 uniform filter, and then computing the partial derivative along
   the y-axis of an image that was smoothed with a 31 uniform filter. The
   two partial derivatives are derivatives of different images! Needless to
   say, if we combine these into one gradient vector we obtain a nonsensical
   result (or a poor approximation at best). This means that these filters
   are not rotation invariant.
*/

#include <stdio.h>
#include <math.h>

double LoG( int x, int y, double sigma );
double Gaussian( int x, int y, double sigma );
double Gaussian_x( int x, int y, double sigma );
double Gaussian_y( int x, int y, double sigma );
void printEntry( double t, int a, int b );

main()
{
    int a, b;

    /*
     *  Guideline: n <= ceil( 3 * sigma )
     */
    const double sigma = 2.5;
    const int n = (int)ceil( 3.0 * sigma );
    printf( "sigma = %4.1f\n", sigma );

    /*
     *  Generate derivative of Gaussian filter.
     */
/*
    const double m = -12.0/Gaussian_x( -1, 0, sigma );
    //const double m = -12.0/Gaussian_y( 0, -1, sigma );
    printf( "double vx =\n" );	// horizontal
    for( a = -n; a <= n; a++ ) {
	for( b = -n; b <= n; b++ ) {
	    //int t = (int)(m * Gaussian_y( a, b, sigma ));
	    double t = m * Gaussian_y( a, b, sigma );
	    printEntry( t, a, b );
	}
	printf( "\n" );
    }
    printf( "double vy =\n" );	// vertical
    for( a = -n; a <= n; a++ ) {
	for( b = -n; b <= n; b++ ) {
	    //int t = (int)(m * Gaussian_x( a, b, sigma ));
	    double t = m * Gaussian_x( a, b, sigma );
	    printEntry( t, a, b );
	}
	printf( "\n" );
    }
*/

    /*
     *  Generate Laplacian of Gaussian (LoG) filter.
     */
    //const double m = 40.0/Gaussian( 0, 0, sigma );
    const double m = 40.0/LoG( 0, 0, sigma );
    printf( "double v =\n" );
    for( a = -n; a <= n; a++ ) {
	for( b = -n; b <= n; b++ ) {
	    //int t = (int)(m * Gaussian( a, b, sigma ));
	    //double t = (int)(m * Gaussian( a, b, sigma ));
	    //int t = (int)(m * LoG( a, b, sigma ));
	    double t = m * LoG( a, b, sigma );
	    printEntry( t, a, b );
	}
	printf( "\n" );
    }
/*
*/

}

/*
 *  Laplacian of Gaussian kernel.
 */
double LoG( int x, int y, double sigma )
{
    double e = -(1.0 / (M_PI * sigma * sigma * sigma * sigma))
		* (1.0 - (x*x + y*y) / (2.0 * sigma * sigma))
		* exp( - (double)(x*x + y*y) / (2.0 * sigma * sigma) );

    return( e );
}

/*
 *  Gaussian.
 */
double Gaussian( int x, int y, double sigma )
{
    double e = (1.0 / (2.0 * M_PI * sigma * sigma))
		* exp( - (double)(x*x + y*y) / (2.0 * sigma * sigma) );

    return( e );
}

/*
 *  Derivative of Gaussian with respect to x.
 */
double Gaussian_x( int x, int y, double sigma )
{
    double e = -(1.0 / (2.0 * M_PI * sigma * sigma * sigma * sigma))
		* x
		* exp( - (double)(x*x + y*y) / (2.0 * sigma * sigma) );

    return( e );
}

/*
 *  Derivative of Gaussian with respect to y.
 */
double Gaussian_y( int x, int y, double sigma )
{
    double e = -(1.0 / (2.0 * M_PI * sigma * sigma * sigma * sigma))
		* y
		* exp( - (double)(x*x + y*y) / (2.0 * sigma * sigma) );

    return( e );
}

void printEntry( double t, int a, int b )
{
    if( (a < 0) && (b < 0) ) {
	printf( " %12.8f*f[k%2d*w%2d] +\n", t, a, b );
    }
    else
    if( (a < 0) && (b >= 0) ) {
	printf( " %12.8f*f[k%2d*w+%1d] +\n", t, a, b );
    }
    else
    if( (a >= 0) && (b < 0) ) {
	printf( " %12.8f*f[k+%1d*w%2d] +\n", t, a, b );
    }
    else {
	printf( " %12.8f*f[k+%1d*w+%1d] +\n", t, a, b );
    }
}

